/*################################################################ */
/*  Part of CPYTH package, interface between COMPHEP and PYTHIA   */
/*                   Monte-Carlo generators                       */
/*################################################################ */
/*  author:  A.Sherstnev                                          */
/*  version:  2.0.01                                              */
/*  date:    28.07.2002                                           */
/*################################################################ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "structures.h"
#include "fill.h"
#include "tag_writer.h"
#include "tag_parser.h"
#include "compare.h"

/****************************************************/
/*             PARTICLE NUMBER routine              */
/****************************************************/
static int
particle (char *pname)
{
/* Leptons */
  if (!strcmp (pname, "e1"))
    return 0;
  if (!strcmp (pname, "e"))
    return 0;
  if (!strcmp (pname, "E1"))
    return 1;
  if (!strcmp (pname, "E"))
    return 1;
  if (!strcmp (pname, "e2"))
    return 2;
  if (!strcmp (pname, "m"))
    return 2;
  if (!strcmp (pname, "E2"))
    return 3;
  if (!strcmp (pname, "M"))
    return 3;
  if (!strcmp (pname, "e3"))
    return 4;
  if (!strcmp (pname, "l"))
    return 4;
  if (!strcmp (pname, "E3"))
    return 5;
  if (!strcmp (pname, "L"))
    return 5;
  if (!strcmp (pname, "n1"))
    return 6;
  if (!strcmp (pname, "ne"))
    return 6;
  if (!strcmp (pname, "N1"))
    return 7;
  if (!strcmp (pname, "Ne"))
    return 7;
  if (!strcmp (pname, "n2"))
    return 8;
  if (!strcmp (pname, "nm"))
    return 8;
  if (!strcmp (pname, "N2"))
    return 9;
  if (!strcmp (pname, "Nm"))
    return 9;
  if (!strcmp (pname, "n3"))
    return 10;
  if (!strcmp (pname, "nl"))
    return 10;
  if (!strcmp (pname, "N3"))
    return 11;
  if (!strcmp (pname, "Nl"))
    return 11;
  if (!strcmp (pname, "e4"))
    return 12;
  if (!strcmp (pname, "E4"))
    return 13;
  if (!strcmp (pname, "n4"))
    return 14;
  if (!strcmp (pname, "N4"))
    return 15;
/* Quarks */
  if (!strcmp (pname, "d#"))
    return 16;
  if (!strcmp (pname, "D#"))
    return 17;
  if (!strcmp (pname, "u#"))
    return 18;
  if (!strcmp (pname, "U#"))
    return 19;
  if (!strcmp (pname, "q#"))
    return 20;
  if (!strcmp (pname, "Q#"))
    return 21;
  if (!strcmp (pname, "d"))
    return 22;
  if (!strcmp (pname, "D"))
    return 23;
  if (!strcmp (pname, "u"))
    return 24;
  if (!strcmp (pname, "U"))
    return 25;
  if (!strcmp (pname, "s"))
    return 26;
  if (!strcmp (pname, "S"))
    return 27;
  if (!strcmp (pname, "c"))
    return 28;
  if (!strcmp (pname, "C"))
    return 29;
  if (!strcmp (pname, "b"))
    return 30;
  if (!strcmp (pname, "B"))
    return 31;
  if (!strcmp (pname, "t"))
    return 32;
  if (!strcmp (pname, "T"))
    return 33;
  if (!strcmp (pname, "b4"))
    return 34;
  if (!strcmp (pname, "B4"))
    return 35;
  if (!strcmp (pname, "t4"))
    return 36;
  if (!strcmp (pname, "T4"))
    return 37;
/* Gauge bosons */
  if (!strcmp (pname, "G"))
    return 38;
  if (!strcmp (pname, "g"))
    return 38;
  if (!strcmp (pname, "A"))
    return 39;
  if (!strcmp (pname, "gamma"))
    return 39;
  if (!strcmp (pname, "Z"))
    return 40;
  if (!strcmp (pname, "Z0"))
    return 40;
  if (!strcmp (pname, "W+"))
    return 41;
  if (!strcmp (pname, "W-"))
    return 42;
/* Higgs bosons */
  if (!strcmp (pname, "H"))
    return 43;
  if (!strcmp (pname, "h"))
    return 43;
  if (!strcmp (pname, "h0"))
    return 43;
  if (!strcmp (pname, "H0"))
    return 44;
  if (!strcmp (pname, "A0"))
    return 45;
  if (!strcmp (pname, "H3"))
    return 45;
  if (!strcmp (pname, "H+"))
    return 46;
  if (!strcmp (pname, "H-"))
    return 47;
/* Gravitons */
  if (!strcmp (pname, "Gr"))
    return 48;
  if (!strcmp (pname, "R0"))
    return 49;
  if (!strcmp (pname, "LQ"))
    return 50;
/* SUSY */
  if (!strcmp (pname, "~g "))
    return 51;
  if (!strcmp (pname, "~dL"))
    return 52;
  if (!strcmp (pname, "~DL"))
    return 53;
  if (!strcmp (pname, "~dR"))
    return 54;
  if (!strcmp (pname, "~DR"))
    return 55;
  if (!strcmp (pname, "~uL"))
    return 56;
  if (!strcmp (pname, "~UL"))
    return 57;
  if (!strcmp (pname, "~uR"))
    return 58;
  if (!strcmp (pname, "~UR"))
    return 59;
  if (!strcmp (pname, "~sL"))
    return 60;
  if (!strcmp (pname, "~SL"))
    return 61;
  if (!strcmp (pname, "~sR"))
    return 62;
  if (!strcmp (pname, "~SR"))
    return 63;
  if (!strcmp (pname, "~cL"))
    return 64;
  if (!strcmp (pname, "~CL"))
    return 65;
  if (!strcmp (pname, "~cR"))
    return 66;
  if (!strcmp (pname, "~CR"))
    return 67;
  if (!strcmp (pname, "~b1"))
    return 68;
  if (!strcmp (pname, "~B1"))
    return 69;
  if (!strcmp (pname, "~b2"))
    return 70;
  if (!strcmp (pname, "~B2"))
    return 71;
  if (!strcmp (pname, "~t1"))
    return 72;
  if (!strcmp (pname, "~T1"))
    return 73;
  if (!strcmp (pname, "~t2"))
    return 74;
  if (!strcmp (pname, "~T2"))
    return 75;
  if (!strcmp (pname, "~eL"))
    return 76;
  if (!strcmp (pname, "~EL"))
    return 77;
  if (!strcmp (pname, "~eR"))
    return 78;
  if (!strcmp (pname, "~ER"))
    return 79;
  if (!strcmp (pname, "~ne"))
    return 80;
  if (!strcmp (pname, "~Ne"))
    return 81;
  if (!strcmp (pname, "~mL"))
    return 82;
  if (!strcmp (pname, "~ML"))
    return 83;
  if (!strcmp (pname, "~mR"))
    return 84;
  if (!strcmp (pname, "~MR"))
    return 85;
  if (!strcmp (pname, "~nm"))
    return 86;
  if (!strcmp (pname, "~Nm"))
    return 87;
  if (!strcmp (pname, "~l1"))
    return 88;
  if (!strcmp (pname, "~L1"))
    return 89;
  if (!strcmp (pname, "~l2"))
    return 90;
  if (!strcmp (pname, "~L2"))
    return 91;
  if (!strcmp (pname, "~nl"))
    return 92;
  if (!strcmp (pname, "~Nl"))
    return 93;
  if (!strcmp (pname, "~o1"))
    return 94;
  if (!strcmp (pname, "~o2"))
    return 95;
  if (!strcmp (pname, "~o3"))
    return 96;
  if (!strcmp (pname, "~o4"))
    return 97;
  if (!strcmp (pname, "~1+"))
    return 98;
  if (!strcmp (pname, "~1-"))
    return 99;
  if (!strcmp (pname, "~2+"))
    return 100;
  if (!strcmp (pname, "~2-"))
    return 101;
  if (!strcmp (pname, "~G"))
    return 102;
  if (!strcmp (pname, "~sS"))
    return 103;
  if (!strcmp (pname, "~sP"))
    return 104;
  return -1;
}

/****************************************************/
/* Compare two process using their particle content */
/* 1 - the processes coinside                       */
/* 0 - the processes are different                  */
/****************************************************/
static int
proc (process_info f, process_info s)
{
  int i;
  int temp;
  int fkf[2], skf[2];
  int f_pro[NUMBPARTICLES];
  int s_pro[NUMBPARTICLES];

  for (i = 0; i < NUMBPARTICLES; i++)
    {
      f_pro[i] = 0;
      s_pro[i] = 0;
    }
  if (f.Nparton == s.Nparton)
    {
      for (i = 0; i < f.Nparton; i++)
	{
	  if (f.parton[i].in == 1)
	    fkf[0] = f.parton[i].KF;
	  if (f.parton[i].in == 2)
	    fkf[1] = f.parton[i].KF;
	  if (s.parton[i].in == 1)
	    skf[0] = s.parton[i].KF;
	  if (s.parton[i].in == 2)
	    skf[1] = s.parton[i].KF;
	}
      if ((skf[0] != fkf[0]) || (skf[1] != fkf[1]))
	{
	  return 0;
	}
      for (i = 2; i < f.Nparton; i++)
	{
	  temp = particle (f.parton[i].name);
	  if (temp >= 0)
	    {
	      f_pro[temp] += 1;
	    }
	  else
	    {
	      fprintf (stderr, "There is unkonwn particle '%s' in process %s\n",
		       f.parton[i].name, f.name);
	      exit (4);
	    }
	  temp = particle (s.parton[i].name);
	  if (temp >= 0)
	    {
	      s_pro[temp] += 1;
	    }
	  else
	    {
	      fprintf (stderr, "There is unkonwn particle '%s' in process %s\n",
		       s.parton[i].name, s.name);
	      exit (4);
	    }
	}
      for (i = 0; i < NUMBPARTICLES; i++)
	{
	  if (f_pro[i] != s_pro[i])
	    return 0;
	}
      return 1;
    }
  return 0;
}

int
compare_processes (int nf, tags ** tagbase, proc_pos * pr_pos)
{
  int i, j, k, n;
  int subtot;
  int result;
  int Nproc;
  string_comnd com;
  process_info *pr;

  subtot = 0;
  pr = malloc (sizeof (process_info));
  if (com.name == NULL || com.value == NULL || pr == NULL)
    {
      fprintf (stderr, "Error: problems with memory allocation");
      return 0;
    }
  for (i = 0; i < nf; i++)
    {
      Nproc = number_proc (tagbase[i]);
      pr = realloc (pr, (subtot + Nproc) * sizeof (process_info));
      for (j = 0; j < Nproc; j++)
	{
	  strcpy (com.name, "ID");
	  sprintf (com.value, "%i", j + 1);
	  n = get_tag_with_exactcom (0, tagbase[i], "process", &com);
	  if (n != -1)
	    {
	      pr[j + subtot].Nparton = get_ival (0, "Nparton", tagbase[i]->tag[n]);
	      str_cpy (pr[j + subtot].name, get_cval (0, "name", tagbase[i]->tag[n]));
	    }
	  n = -1;
	  strcpy (com.name, "IDprocess");
	  for (k = 0; k < pr[j + subtot].Nparton; k++)
	    {
	      n = get_tag_with_exactcom (n + 1, tagbase[i], "parton", &com);
	      if (n != 1)
		{
		  pr[j + subtot].parton[k].in = get_ival (0, "in", tagbase[i]->tag[n]);
		  pr[j + subtot].parton[k].out = get_ival (0, "out", tagbase[i]->tag[n]);
		  pr[j + subtot].parton[k].in = get_ival (0, "in", tagbase[i]->tag[n]);
		  pr[j + subtot].parton[k].out = get_ival (0, "out", tagbase[i]->tag[n]);
		  pr[j + subtot].parton[k].KF = get_ival (0, "KF", tagbase[i]->tag[n]);
		  str_cpy (pr[j + subtot].parton[k].name, get_cval (0, "name", tagbase[i]->tag[n]));
		}
	    }

	  pr_pos[j + subtot].nfile = i;
	  n = get_tag_with_exactcom (0, tagbase[i], "n_event", &com);
	  if (n != -1)
	    pr_pos[j + subtot].Nevents = get_ival (0, "N", tagbase[i]->tag[n]);
	}
      subtot += Nproc;
    }

  for (i = 0; i < subtot - 1; i++)
    {
      for (j = i + 1; j < subtot; j++)
	{
	  result = proc (pr[i], pr[j]);
	  if (result)
	    {
	      fprintf (stderr, "Warning: You are trying to mix identical processes!\n");
	      fprintf (stderr, "         their numbers %i and %i\n", i + 1, j + 1);
	    }
	}
    }
  free (pr);
  return 0;
}

static int
proto_crucial (elementary_tag f, elementary_tag s)
{
  return 1;
}

static int
ele_crucial (elementary_tag f, elementary_tag s)
{
  return 1;
}

static int
photo_crucial (elementary_tag f, elementary_tag s)
{
  return 1;
}

int
compare_extra_info (int nf, tags ** t, process_ * pr)
{
  int l, i, j, k;
  int res, n;
  string_comnd com;
  elementary_tag *beam[nf][2];
  elementary_tag *strfun[nf][2];

  if (com.name == NULL || com.value == NULL)
    {
      fprintf (stderr, "Error: problems with memory allocation");
      return 0;
    }
  for (j = 0; j < nf; j++)
    {
      for (i = 0; i < 2; i++)
	{
	  strcpy (com.name, "ID");
	  sprintf (com.value, "%i", i + 1);
	  n = get_tag_with_exactcom (0, t[j], "beam", &com);
	  if (n != -1)
	    {
	      beam[j][i] = t[j]->tag[n];
	    }
	  strcpy (com.name, "IDbeam");
	  n = get_tag_with_exactcom (0, t[j], "strfun", &com);
	  if (n != -1)
	    {
	      strfun[j][i] = t[j]->tag[n];
	    }
	}
    }

  for (l = 0; l < 2; l++)
    {
      if (abs (pr->proc_info.IDBeamUP[l]) == 2212)
	{
	  for (i = 0; i < 2; i++)
	    {
	      for (j = 0; j < nf - 1; j++)
		{
		  for (k = j + 1; k < nf; k++)
		    {
		      res = proto_crucial (*strfun[j][i], *strfun[k][i]);
		      if (!res)
			{
			  fprintf (stderr, "\nError: proton extra commands\n");
			  exit (2);
			}
		    }
		}
	    }
	  return 1;
	}
      else if (abs (pr->proc_info.IDBeamUP[l]) == 11)
	{
	  for (i = 0; i < 2; i++)
	    {
	      for (j = 0; j < nf - 1; j++)
		{
		  for (k = j + 1; k < nf; k++)
		    {
		      res = ele_crucial (*strfun[j][i], *strfun[k][i]);
		      if (!res)
			{
			  fprintf (stderr, "\nError: electron extra commands\n");
			  exit (2);
			}
		    }
		}
	    }
	  return 1;
	}
      else if (abs (pr->proc_info.IDBeamUP[l]) == 22)
	{
	  for (i = 0; i < 2; i++)
	    {
	      for (j = 0; j < nf - 1; j++)
		{
		  for (k = j + 1; k < nf; k++)
		    {
		      res = photo_crucial (*strfun[j][i], *strfun[k][i]);
		      if (!res)
			{
			  fprintf (stderr, "\nError: photon extra commands\n");
			  exit (2);
			}
		    }
		}
	    }
	  return 1;
	}
      else
	{
	  return 0;
	}
    }
  return 0;
}
