/* 
   Copyright (C) 2002, Alexander Sherstnev 
   $Id $   
   $Log $
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include "tag_reader.h"
#include "tag_writer.h"
#include "tag_parser.h"
#include "fill.h"
#include "validity.h"
#include "structures.h"
#include "compare.h"
#include "mix.h"

int 
main (int argc, char **argv)
{
  int nf, jf, t, tt, nx;
  int nSubtot;
  int nProc;
  int *shift;
  long i, l;
  long nRectot;
  long maxnRec;
  long xxtim;
  long num;
  long **pos;
  long fpos;
  long *pos0;
  long *nRec;
  long *real_nRec;
  long *nRecRest;
  double *ri, xrn;
  double sigmatot = 0.0;
  char *outFileName;
  char *prtFileName;
  char buff[1000];
  char buff1[1000];
  char **map;
  char **in_filename;

  FILE **inFile;
  FILE *outFile;
  FILE *prtFile;
  process_ theProcessUP;
  tags **tbase;
  proc_pos *proc_position;

/****************************************************************/
/*     initialize file names                                    */

  outFileName = "Mixed.PEV";
  prtFileName = "Prt.PEV";

/****************************************************************/
/*  allocate the memory space for work dimensions */
  nf = argc - 1;
  in_filename = malloc (nf * sizeof (char *));
  tbase = malloc (nf * sizeof (tags *));
  inFile = malloc (nf * sizeof (FILE *));
  pos0 = malloc (nf * sizeof (long));
  shift = malloc (nf * sizeof (int));

/****************************************************************/
/* open output files for writing                                */
  prtFile = fopen (prtFileName, "w");

/****************************************************************/
/* write header of protocol file                                */
  fprintf (prtFile, " Mix program, version 2.1.00");
  time (&xxtim);
  fprintf (prtFile, "Time: %s", ctime (&xxtim));
  fprintf (prtFile, "Files to mix and randomize:\n");

/****************************************************************/
/* check arguments - input event files                          */
  if (argc < 2)
    {
      fprintf (stderr, "Enter names of files with events as parameters. \n");
      return 1;
    }

  for (jf = 0; jf < nf; jf++)
    {
      in_filename[jf] = argv[jf + 1];
      inFile[jf] = fopen (in_filename[jf], "r");
      if (!inFile[jf])
	{
	  fprintf (stderr, "File not found: %s\n", argv[jf + 1]);
	  fprintf (prtFile, "File not found: %s\n", argv[jf + 1]);
	  return 2;
	}
      else
	{
	  fprintf (prtFile, "%s\n", argv[jf + 1]);
	}
    }

  for (i = 0; i < nf; i++)
    {
      tbase[i] = init_cap (1);
      cup_reader (inFile[i], tbase[i]);
      syntax_validity (tbase[i]);
      structure_validity (tbase[i]);
      physics_validity (tbase[i]);
      pos0[i] = ftell (inFile[i]) + 1;	/* position of the 1st event beginning */
    }

  fill_LH_structures (nf, tbase, &theProcessUP, in_filename);
  compare_extra_info (nf, tbase, &theProcessUP);
  nSubtot = theProcessUP.proc_info.NprocRUP;
  proc_position = malloc (nSubtot * sizeof (proc_pos));
  compare_processes (nf, tbase, proc_position);

/****************************************************************/
/*   allocate the memory space for dimensions                   */

  nRec = malloc (nSubtot * sizeof (long));
  real_nRec = malloc (nSubtot * sizeof (long));
  nRecRest = malloc (nSubtot * sizeof (long));
  ri = malloc (nSubtot * sizeof (double));
  pos = malloc (nSubtot * sizeof (long *));
  map = malloc (nSubtot * sizeof (char *));

  nRectot = 0;
  maxnRec = 0;
  for (i = 0; i < nSubtot; i++)
    real_nRec[i] = 0;

  t = 0;
  for (i = 0; i < nf; i++)
    {
      tt = get_tag (0, tbase[i], "total");
      if (tt != -1)
	tt = get_ival (0, "Nproc", tbase[i]->tag[tt]);
      else
	{
	  fprintf (stderr, "Error: the file %s does not contain the total tag.\n", in_filename[i]);
	  exit (2);
	}

      fseek (inFile[i], pos0[i], SEEK_SET);
      while (1)
	{
	  if (!fgets (buff, 1000, inFile[i]))
	    break;
	  sscanf (buff, "%i", &nProc);
	  num = t + nProc - 1;
	  if (num < nSubtot)
	    real_nRec[num]++;
	  else
	    {
	      fprintf (stderr, "Error: process number of event exceed max process number in file %s.\n", in_filename[i]);
	      exit (2);
	    }
	}
      t += tt;
    }

  sigmatot = 0.0;
  for (i = 0; i < nSubtot; i++)
    {
      nRec[i] = proc_position[i].Nevents;
      if (real_nRec[i] != nRec[i])
	{
	  fprintf (stderr, "Warning! real number of events in file %s (%li) does not coinside with header info (%li).\n",
		   in_filename[i], real_nRec[i], nRec[i]);
	  nRec[i] = real_nRec[i];
// Nuzhno zamenit' nepravil'niy Nevents dlya etogo processa.      
	}
      nRecRest[i] = nRec[i];
      nRectot = nRectot + nRec[i];
      if (nRec[i] > maxnRec)
	{
	  maxnRec = nRec[i];
	}
      pos[i] = malloc (nRec[i] * sizeof (long));
      map[i] = malloc (nRec[i] * sizeof (char));
      sigmatot += theProcessUP.proc_info.crossecUP[i];
    }

  t = 0;
  for (i = 0; i < nSubtot; i++)
    real_nRec[i] = 0;
  for (i = 0; i < nf; i++)
    {
      tt = get_tag (0, tbase[i], "total");
      if (tt != -1)
	tt = get_ival (0, "Nproc", tbase[i]->tag[tt]);
      else
	{
	  fprintf (stderr, "Error: the file %s des not contain the total tag.\n", in_filename[i]);
	  exit (2);
	}
      shift[i] = t;

      fseek (inFile[i], pos0[i], SEEK_SET);
      while (1)
	{
	  fpos = ftell (inFile[i]);
	  if (!fgets (buff, 1000, inFile[i]))
	    break;
	  sscanf (buff, "%i", &nProc);
	  num = t + nProc - 1;
	  pos[num][real_nRec[num]] = fpos;
	  map[num][real_nRec[num]] = 0;
	  real_nRec[num]++;
	}
      t += tt;
    }

/******************************************************************/
/* evaluate the map of subprocesses weights in the interval [0,1] */
  xrn = 0;
  for (i = 0; i < nSubtot; i++)
    {
      ri[i] = xrn + theProcessUP.proc_info.crossecUP[i] / sigmatot;
      xrn = ri[i];
    }

/****************************************************************/
/*         mix events from the files                            */
  outFile = fopen (outFileName, "w");


  write_cap (outFile, tbase, theProcessUP);

  for (i = 0; i < nRectot; i++)
    {
      xrn = drand48 ();
      jf = 0;
      while (xrn > ri[jf])
	{
	  jf++;
	}
      if (nRecRest[jf] == 0)
	{
	  fprintf (stderr,
		   " Mixed/randomized events (file %s exhausted): %ld\n", argv[proc_position[jf].nfile + 1], i);
	  fprintf (prtFile,
		   " Mixed/randomized events (file %s exhausted): %ld\n", argv[proc_position[jf].nfile + 1], i);
	  fclose (outFile);
	  fclose (prtFile);
	  final_write_cap (outFileName, nRecRest, i);
	  free (pos);
	  free (map);
	  free (proc_position);
	  free (tbase);
	  free (inFile);
	  free (pos0);
	  free (shift);
	  free (nRec);
	  free (nRecRest);
	  free (ri);
	  return 1;
	}
//    ll=drand48();
      //    l=ll*nRec[jf];
      l = 0;
      while (map[jf][l])
	{
	  l++;
	  if (l == nRec[jf])
	    {
	      l = 0;
	    }
	}
      fseek (inFile[proc_position[jf].nfile], pos[jf][l], SEEK_SET);
      map[jf][l] = 1;
      fgets (buff, 1000, inFile[proc_position[jf].nfile]);
      sscanf (buff, "%i:%[^\n]", &nx, buff1);
      fprintf (outFile, " %i:%s\n", (int) (shift[proc_position[jf].nfile] + nx), buff1);
      nRecRest[jf] -= 1;
    }

  fprintf (stderr, " Mixed/randomized events: %ld\n", nRectot);
  fprintf (prtFile, " Mixed/randomized events: %ld\n", nRectot);
  fclose (outFile);
  fclose (prtFile);
  final_write_cap (outFileName, nRecRest, i);
  free (pos);
  free (map);
  free (proc_position);
  free (tbase);
  free (inFile);
  free (pos0);
  free (shift);
  free (nRec);
  free (nRecRest);
  free (ri);
  return 1;
}
