/*#######################################################################*/
/*    Program for mix and randomize events from several channels (files) */
/*                      in one event flow                                */
/*    Input files: in CompHEP v.41   or PEVLIB v.10 formats              */
/*#######################################################################*/
/*  author:  V.A.Ilyin                                                   */
/*  date:    04.06.2000                                                  */
/*#######################################################################*/

#include<stdio.h>
#include<stdlib.h>
#include <time.h>
char *xpevlib,xpevlib2[72],xpevlib3[72],*xdefis,*xevents,
     *xcomphep,*xcomphep1,xtmpfile[127],*xnsub,*xcross,*xmix;
char *outFileName,*prtFileName;

int main(int argc,char** argv)
{ char buff[1000],buff1[1000],word[100],xstr[72];
  int nf,jf,nx,nSubtot,nlen;
  long i,l,ipos,nRectot,maxnRec,posTotEv,xxtim;
  long **pos,*fpos,*pos0,*nsub,*nRec,*nRecRest;
  char **map,*fmap;
  double *sigma,*ri,sigmatot,xrn;
  FILE **inFile,*outFile,*prtFile;
  
/****************************************************************/
/*     initialize global variables                              */

  xcomphep = "#CompHEP_41.10\n";
  xcomphep1 = "#CompHEP_41.20\n";
  xpevlib = "#PEVLIB_v.1.0";
  strcpy(xpevlib2,xpevlib);
  strcpy(xpevlib3,xpevlib);
  strcat(xpevlib2," =============================================\n");
  strcat(xpevlib3,"\n");
  xdefis =  "#----------------------------------------------------------\n";
  xevents = "#Nproc ================== Events ==========================\n";
  xnsub = "#Number_of_subprocesses";
  xcross = "#Total_cross_section_(pb)";
  xmix = "#Events_mixed_and_randomized";
  outFileName="Mixed.PEV";
  prtFileName="Prt.PEV";

/****************************************************************/
/*  allocate the memory space for work dimensions */
  nf=argc-1;
  pos = malloc(nf*sizeof(long*));
  map = malloc(nf*sizeof(char*));
  inFile = malloc(nf*sizeof(FILE*));
  outFile = malloc(nf*sizeof(FILE*));
  prtFile = malloc(nf*sizeof(FILE*));
  sigma = malloc(nf*sizeof(double));
  ri= malloc(nf*sizeof(double));
  pos0 = malloc(nf*sizeof(long*));
  nsub = malloc(nf*sizeof(long*));
  nRec = malloc(nf*sizeof(long*));
  nRecRest = malloc(nf*sizeof(long*));

/****************************************************************/
/* open output files for writing                                */
  outFile=fopen(outFileName,"w");
  prtFile=fopen(prtFileName,"w");

/* write header to protocol file                                */
  fprintf(prtFile,"%s",xcomphep);
  fprintf(prtFile,"%s",xpevlib3);
  time(&xxtim);
  fprintf(prtFile,"Time: %s",ctime(&xxtim));
  fprintf(prtFile,"Files to mix and randomize:\n");

/****************************************************************/
/* check arguments - input event files */
  if(argc < 2)
  { fprintf(stderr,"Enter names of files with events as parameters. \n");
    return 1;
  }
  for(jf=0;jf<nf;jf++)
  { inFile[jf]=fopen(argv[jf+1],"r");
    if(!inFile[jf])
    { fprintf(stderr,"File not found: %s\n",argv[jf+1]);
      fprintf(prtFile,"File not found: %s\n",argv[jf+1]);
      return 2;
    }
    else fprintf(prtFile,"%s\n",argv[jf+1]);
  }

/****************************************************************/
/*        check PEVLIB headers of event files                   */
  for (jf=0;jf<nf;jf++)
  { fgets(buff,1000,inFile[jf]);
    sscanf(buff,"%s",word);
    if(strcmp(word,"#CompHEP")==0)
    { fclose(inFile[jf]);
      toPEVLIB(argv[jf+1]);
      inFile[jf]=fopen(xtmpfile,"r");
    }
  }

/****************************************************************/
/*        read-write headers of event files                     */
  sigmatot = 0;
  nSubtot = 0;
  sprintf(buff,xpevlib2);
  fputs(buff,outFile);
  for (jf=0;jf<nf;jf++)
  {while(1) {fgets(buff,1000,inFile[jf]);
             sscanf(buff,"%s",word);
             if(strcmp(word,xnsub)==0) break;
             if(strcmp(word,xpevlib)) fputs(buff,outFile);
            }
     strcpy(xstr,xnsub);
     strcat(xstr," = %ld");
     sscanf(buff,xstr,nsub+jf);
   fgets(buff,1000,inFile[jf]);
     strcpy(xstr,xcross);
     strcat(xstr," = %lf");
     sscanf(buff,xstr,&sigma[jf]);
   fgets(buff,1000,inFile[jf]);
     strcpy(xstr,xmix);
     strcat(xstr," = %ld");
     sscanf(buff,xstr,&nRec[jf]);
     fgets(buff,1000,inFile[jf]);
   pos0[jf]=ftell(inFile[jf]);   /* position of the 1st event beginning */
   nSubtot = nSubtot+nsub[jf];
   nsub[jf] = nSubtot-nsub[jf];
   sigmatot = sigmatot+sigma[jf];
  }
   
/****************************************************************/
/*   allocate the memory space for dimensions                   */
  nRectot = 0;
  maxnRec = 0;
  for (jf=0;jf<nf;jf++)
  {
   i=0;
   while(1)
   {if(fgets(buff,1000,inFile[jf])==0) break;
    i++;
   }
   nRec[jf]=i;
   nRecRest[jf]=i;
   nRectot = nRectot+nRec[jf];
   if(nRec[jf]>maxnRec) maxnRec=nRec[jf];
  }
  
  for (jf=0;jf<nf;jf++)
  {
   fpos=malloc(maxnRec*sizeof(long));
   fmap=malloc(maxnRec*sizeof(char));
   fseek(inFile[jf],pos0[jf],SEEK_SET);
   i=0;
   while(1)
   {fpos[i]=ftell(inFile[jf]);
    fmap[i]=0;
    if(fgets(buff,1000,inFile[jf])==0) break;
    i++;
   }
   pos[jf]=fpos;
   map[jf]=fmap;
  }

/****************************************************************/
/*     write general info in header                             */
  sprintf(buff,"%s = %d\n",xnsub,nSubtot);
  fputs(buff,outFile);
  sprintf(buff,"%s = %E\n",xcross,sigmatot);
  fputs(buff,outFile);
  sprintf(buff,"%s = %ld\n",xmix,nRectot);
  posTotEv=ftell(outFile);
  fputs(buff,outFile);
  sprintf(buff,xevents);
  fputs(buff,outFile);

/******************************************************************/
/* evaluate the map of subprocesses weights in the interval [0,1] */
   xrn = 0;
   for (jf=0;jf<nf;jf++) {ri[jf]=xrn+sigma[jf]/sigmatot; xrn=ri[jf];}

/****************************************************************/
/*         mix events from the files                            */
  for(i=0;i<nRectot;i++)
  { xrn = drand48();
    jf = 0; while (xrn>ri[jf]) jf++;
    if (nRecRest[jf]==0)
     { fprintf(stderr,
         " Mixed/randomized events (file %s exhausted): %ld\n",argv[jf+1],i);
       fprintf(prtFile,
         " Mixed/randomized events (file %s exhausted): %ld\n",argv[jf+1],i);
       nx=TotEvPut(outFile,prtFile,posTotEv,i);
       return nx;
     }
    l=drand48()*nRec[jf];
    while(map[jf][l]) {l++; if (l==nRec[jf]) l=0;}
    fseek(inFile[jf],pos[jf][l],SEEK_SET);
    map[jf][l]=1;
    ipos = ftell(inFile[jf]);
    fgets(buff,1000,inFile[jf]);
    sscanf(buff,"%ld",&nx);
    sscanf(buff,"%s",word);
    nlen = strlen(word);
    fseek(inFile[jf],ipos,SEEK_SET);
    fgets(word,nlen+6,inFile[jf]);
    fgets(buff1,1000,inFile[jf]);
    sprintf(buff," %-5li%s",nsub[jf]+nx,buff1); 
    fputs(buff,outFile);
    nRecRest[jf]=nRecRest[jf]-1;
  }

  fprintf(stderr," Mixed/randomized events: %ld\n",nRectot);
  fprintf(prtFile," Mixed/randomized events: %ld\n",nRectot);
       nx=TotEvPut(outFile,prtFile,posTotEv,nRectot);
  return nx;
}


/*############################################################*/
/*                Functions                                   */
/*############################################################*/

/******************************************************************/
/*   transform events file of CompHEP format to the PEVLIB format */
int TotEvPut(outFile,prtFile,posTotEv,TotEv)
FILE *outFile,*prtFile;
long posTotEv,TotEv;
{
   char buff[100],xfmt[20];
   int i,ilen,nlen;
   fclose(outFile);
   outFile=fopen(outFileName,"r+");
   fseek(outFile,posTotEv,SEEK_SET);
   fgets(buff,1000,outFile);
   ilen=strlen(buff)-32;
   sprintf(buff,"%ld",TotEv);
   nlen=strlen(buff); 
   if(ilen-nlen < 0)
     { fprintf(stderr,"Error: mixed events more than sum of input events.\n");
       fprintf(prtFile,"Error: mixed events more than sum of input events.\n");
      return -1;
     }
   sprintf(xfmt,"%s%d%s","%s = %-",ilen,"ld");
   sprintf(buff,xfmt,xmix,TotEv);
   fseek(outFile,posTotEv,SEEK_SET);
   fputs(buff,outFile); 
   fclose(outFile);
   fclose(prtFile);
   return 0;
}

/******************************************************************/
/*   transform events file of CompHEP format to the PEVLIB format */
int toPEVLIB(ifile)
char *ifile;
{
char buff[1000],buff1[30],word[100],xstr[72];
char word_1[8],word_2[7],word_3[5];
int i,nweight,version,while_exit;
long nRec;
double sigma;
FILE *inFile,*outFile;

inFile=fopen(ifile,"r");
  strcpy(xtmpfile,ifile);
  strcat(xtmpfile,"___");
outFile=fopen(xtmpfile,"w");

sprintf(buff,xpevlib2);
fputs(buff,outFile);


fgets(buff1,30,inFile);
sscanf(buff1,"%s %s %s",word_1,word_2,word_3);
fputs(buff1,outFile);

  if (strcmp(word_1,"#CompHEP")==0)
   {
    if (strcmp(word_3,"41.10")==0) version = 4110;
    if (strcmp(word_3,"41.20")==0) version = 4120;
   }  
  else
  {fprintf(stderr,"Uknown CompHEP version!!!");return 1;}
  
/****************************************************************/
/*        read-write of header                                  */
if(version == 4110)
  {
   fprintf(stderr,"URA 1 !!!\n");
   for(i=0;i<9;i++)
   { 
    fgets(buff,1000,inFile);
    sscanf(buff,"%s",word);
    if(strcmp(word,"#Cross_section(Width)")==0)
       sscanf(buff,"#Cross_section(Width) %lf",&sigma);
    if(strcmp(word,"#Number_of_events")==0)
       sscanf(buff,"#Number_of_events  %li",&nRec);
    fputs(buff,outFile);
   }

    sprintf(buff,xdefis);
      fputs(buff,outFile);
    strcpy(xstr,xnsub);
    strcat(xstr," = %d\n");
    sprintf(buff,xstr,1);
      fputs(buff,outFile);
    strcpy(xstr,xcross);
    strcat(xstr," = %E\n");
    sprintf(buff,xstr,sigma);
      fputs(buff,outFile);
    strcpy(xstr,xmix);
    strcat(xstr," = %ld\n");
    sprintf(buff,xstr,nRec);
      fputs(buff,outFile);
    sprintf(buff,xevents);
      fputs(buff,outFile);

    fgets(buff,1000,inFile); 
  }

if(version == 4120)
  {
   while_exit=0;
   while(while_exit==0)
    { 
     fgets(buff,1000,inFile);
     sscanf(buff,"%s",word);  
     if(strcmp(word,"#Total_cross_section_(pb)")==0)
        sscanf(buff,"#Total_cross_section_(pb) = %f",&sigma);
     if(strcmp(word,"#Events_mixed_and_randomized")==0)
        sscanf(buff,"#Events_mixed_and_randomized = %i",&nRec);
     if(strcmp(word,"#Nproc")==0) while_exit = 1;
     fputs(buff,outFile);
     }
  }

/****************************************************************/
/*         rewriting of events                                  */
while(fscanf(inFile,"%d",&nweight) == 1)
  {  fgets(buff,1000,inFile);
     while (nweight>0)
       {  
          if(version == 4110)  sprintf(buff1," 1  %s",buff);
          if(version == 4120)  sprintf(buff1," 1%s",buff);
          fputs(buff1,outFile);
          nweight--;
       }
  }

fclose(outFile);
return 0;
}
